/*
 milix Class Library v1.1
 Binary Search Tree

 FILE  : bstree.h
 AUTHOR: Anastasios Milikas (aka 'milix'). e-mail: milix@milix.net

 Copyright (c) 1999-2003 milix
 Permission is hereby granted, free of charge, to any person obtaining a copy of
 this software and associated documentation files (the "Software"), to deal in
 the Software without restriction, including without limitation the rights to use,
 copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
 Software, and to permit persons to whom the Software is furnished to do so,
 subject to the following conditions:

 The above copyright notice and this permission notice shall be included in all
 copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

 Overview
 ~~~~~~~~
 Binary search tree definition.
 Classes defined here: bstree, bstraveller.
*/

#ifndef _mcl_bstree_h
#define _mcl_bstree_h

// BEGIN NAMESPACE MCL
namespace mcl {

/////////////////////////////////////////////////
// bstree
// ~~~~~~
// Binary Search Tree with balance method.
// If memory pool is OnDisk a file handle must be specified
// before any tree operation (init, insert, find etc.). The
// file must be opened (handle different than -1).

typedef unsigned int node_t;
// A node structure started at node_t (32bit) addr address is
//   4 bytes: left branch
//   4 bytes: right branch
//   N bytes: data

class bstree
{
 public:
	// Memory pool
	enum { OnMem, OnDisk };

	// Constructors, destructor
	bstree();
	bstree(unsigned datasz, unsigned capacity=32, int mpool=OnMem, int fh=-1);
	virtual ~bstree();

	// properties
	void     set_datasize(unsigned datasz) { m_datasz = datasz; }
	unsigned data_size() const { return m_datasz; }
	unsigned capacity() const { return m_capacity; }
	unsigned length() { return m_count; }
	unsigned size() { return length(); }
	void     set_memory(int mpool, int fh=-1);
	int      memory() const { return m_mpool; }
	void*    workspace() { return m_space; }
	char*    buffer() { return m_buf; }

    // each time the tree needs to be expanded, reserve is called with
    // that amount of items.
	void     set_reserve_amount(long amount) { m_amount = amount; }
	
	// reserve pre-allocates memory for 'capacity' items
	int      reserve(unsigned capacity);

	// tree operations
	node_t   insert(const void* item);
	int      erase(node_t node, node_t parent);
    int      clear();
	node_t   find_first(const void* key, node_t* parent=(node_t*)0);
	node_t   find_next(node_t node, const void* key);

	// node & node data
	void     set_data(node_t node, const void* data);
	void*    data(node_t node);
	node_t   left(node_t node);
	node_t   right(node_t node);
	void     set_left(node_t node, node_t point);
	void     set_right(node_t node, node_t point);

	// linear-time, zero space overhead binary tree rebalancing
	void     balance();

	// filling
	int      creat(int fh=-1);
	int      read(int fh=-1);
	int      write(int fh=-1);

 protected:
	// pure private virtual method compare(). Override this in every
	// class derived from bstree to specify the way the keys compared.
	virtual  int compare(const void*, const void*) = 0;

	// private methods
	void     set_root(node_t root);
	node_t   root() const;
	void     set_first_erased(node_t first);
	node_t   first_erased() const;
	long     lref(node_t node) { return node; }
	long     rref(node_t node) { return node+sizeof(long); }
	unsigned estim_space_size();
	node_t   new_node();
	void     del_node(node_t node);
	node_t   left_most(node_t node);
	node_t   right_most(node_t node);

	// private data
	int      m_mpool;
	unsigned m_count;
	unsigned m_datasz;
	unsigned m_capacity;
	long     m_amount;
	void*    m_space;
	char*    m_buf;
};

/////////////////////////////////////////////////
// BSTreeTraveller
// ~~~~~~~~~~~~~~~
// Travelling the nodes of a binary search tree.
//
class bstraveller : protected bstree
{
 public:
	// Ways to visit tree's nodes
	enum { TravelLeft, TravelRight, TravelCurrent };

 public:
	// Constructor.
	bstraveller(bstree*);
	// Destructor, prevent bstree's deallocations
	virtual ~bstraveller() { m_space = (void*)0; m_buf = (char*)0; }

	// Member functions
	int travel(int way);

 protected:
	// Private virtual methods, always override
	virtual int visitor(void* data);

 protected:
	// Private member functions
	virtual int compare(const void*, const void*);
	int _travel_nodes(node_t, int way);
};

// END NAMESPACE MCL
};

#endif   // _mcl_bstree_h
