/*
 milix Class Library v1.1
 Binary Tree

 FILE  : btree.h
 AUTHOR: Anastasios Milikas (aka 'milix'). e-mail: milix@milix.net

 Copyright (c) 1999-2004 by milix
 Permission is hereby granted, free of charge, to any person obtaining a copy of
 this software and associated documentation files (the "Software"), to deal in
 the Software without restriction, including without limitation the rights to use,
 copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
 Software, and to permit persons to whom the Software is furnished to do so,
 subject to the following conditions:

 The above copyright notice and this permission notice shall be included in all
 copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

 Overview
 ~~~~~~~~
 Binary tree definitions.
 Classes defined here: btree, btraveller.
*/

#ifndef _mcl_btree_h
#define _mcl_btree_h

// BEGIN NAMESPACE MCL
namespace mcl {

/////////////////////////////////////////////////
// btree
// ~~~~~
// Binary tree of void* items.
//
class btree
{
 public:
	// Node type
	struct node_t {
		node_t*  left;
		node_t*  right;
		node_t*  parent;
		char*    data;
		unsigned depth;

		bool     islt() const { return bool(parent && parent->left == this); }
		bool     isrt() const { return bool(parent && parent->right == this); }
		unsigned lnodes();
		unsigned rnodes();
		node_t*  left_most_parent();
		node_t*  right_most_parent();
		node_t*  leftmost();
		node_t*  rightmost();
		unsigned parent_distance(node_t* node);
		unsigned left_distance(node_t* node);
		unsigned right_distance(node_t* node);
	};

 public:
	// Constructors. There is no default destructor.
	// To destruct a tree call the member function destroy().
	btree();
	btree(const btree* src);
	virtual void destroy();

	// Member functions
	unsigned&     sizeOf() { return sz; }
	unsigned      length() const { return count; }
	unsigned      size() const { return count; }
	bool          isempty() { return bool(count == 0); }
	node_t*       getroot() { return root; }
	node_t*       current() { return cur; }
	void          set_current(node_t* node) { cur = node; }
	void*         node_data() { return cur ? cur->data : 0; }
	void          clear();
	void          copy(const btree* src);

	// Virtual methods
	virtual bool  walkroot();
	virtual bool  walkleft();
	virtual bool  walkright();
	virtual bool  walkback();
	virtual int   addleft(void* item);
	virtual int   addright(void* item);
	virtual void  deleteleft();
	virtual void  deleteright();
	virtual bool  findline(void* line, unsigned n);
	virtual int   insertline(void* line, unsigned n);
	virtual int   deleteline(void* line, unsigned n);
	virtual int   promoteline(void* line, unsigned n);
	virtual int   demoteline(void* line, unsigned n);

	// Filing
	virtual int   read(int fh);
	virtual int   write(int fh);

 protected:
	// Private data members
	node_t*  root;
	node_t*  cur;
	unsigned count;
	unsigned sz;

	// Override release() to handle the release of memory occupied by
	// the objects inserted in the tree.
	virtual void release(void* data)=0;

	// Override setNodeData() to specify a way other than memcpy() to
	// set the node's data
	virtual void set_node_data(node_t*, const void*);

	// Override read_node_data() and write_node_data() to specify a new
	// store method for the object inserted in the tree.
	virtual int read_node_data(int fh, void*);
	virtual int write_node_data(int fh, void*);
	
	// Private methods
	void        kill(node_t*);
	int         write_nodes(node_t*, int fh);
	node_t*     read_nodes(int fh);
	node_t*     copy_nodes(node_t* src);
};


/////////////////////////////////////////////////
// btraveller
// ~~~~~~~~~~~~~~
// Travelling a binary tree.
//
class btraveller: protected btree
{
 public:
	// Ways to visit tree's nodes
	enum { PreOrder, InOrder, PostOrder, UserBase };

	// public data members
	node_t* root;
	node_t* current;

 public:
	// Constructor
	btraveller(btree*);
	virtual void destroy() {}

	// Member functions
	int travel(int way=PreOrder) { return travel_nodes(root, way); }

 protected:
	// Private virtual method travel_nodes, always override
	virtual int travel_nodes(node_t*, int way)=0;

	// Private member functions
	void release(void*) {}
};

// END NAMESPACE MCL
};

#endif   // _mcl_btree_h
