/*
 milix Class Library v1.1
 Date/time

 FILE  : detetime.h
 AUTHOR: Anastasios Milikas (aka 'milix'). e-mail: milix@milix.net

 Copyright (c) 1999-2003 milix
 Permission is hereby granted, free of charge, to any person obtaining a copy of
 this software and associated documentation files (the "Software"), to deal in
 the Software without restriction, including without limitation the rights to use,
 copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
 Software, and to permit persons to whom the Software is furnished to do so,
 subject to the following conditions:

 The above copyright notice and this permission notice shall be included in all
 copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

 Overview
 ~~~~~~~~
 Date/time classes
*/

#ifndef _mcl_datetime_h
#define _mcl_datetime_h

// include time.h
#if !defined(__TIME_H)
#include <time.h>
#endif

// BEGIN NAMESPACE MCL
namespace mcl {

/////////////////////////////////////////////////
// tdate
// Month 1..12, Day 1..31.
// The first legal date is 01-01-1900

class tdate
{
 public:
    // constructors
    tdate();
    tdate(int year, int mon=1, int mday=1);
    tdate(const tdate& date);
    tdate(const time_t tt);
    
    // member function
    bool     isnull() const { return m_pd == 0; }
    tdate&   mkdate(int year, int mon=1, int mday=1);
    tdate&   current() { return mkdate(0); }
    int      mday() const { return (int)dref(); }
    int      month() const { return (int)mref(); }
    int      year() const { return (int)yref(); }
    int      wday() const;
    int      yday() const;
    int      ydays() const { return (isleap() ? 366:365); }
    int      mdays() const;
    tdate&   nextday();
    tdate&   prevday();
    tdate&   nextmon();
    tdate&   prevmon();
    tdate&   add(int days);
    int      diff(const tdate& date);
    tdate    easter();
    tdate    easter_greece();
    int      compare(const tdate& date) const;
    bool     isleap() const { int y=year(); return (y%4==0 && (y%100!=0||y%400==0)); }
    tdate    str2date(const char* buf);
    char*    date2str(char* buf=(char*)0, int sep='\0');
    unsigned long packed_date() const { return m_pd; }

    // operators
    tdate&   operator = (const tdate& d) { m_pd = d.packed_date(); return *this; }
    bool     operator ==(const tdate& d) const { return compare(d) == 0; }
    bool     operator > (const tdate& d) const { return compare(d) > 0; }
    bool     operator >=(const tdate& d) const { return compare(d) >= 0; }
    bool     operator < (const tdate& d) const { return compare(d) < 0; }
    bool     operator <=(const tdate& d) const { return compare(d) <= 0; }
    tdate&   operator +=(int days) { return add(days); }
    tdate&   operator -=(int days) { return add(-days); }

 protected:
    // private functions
    unsigned char&  dref() const { return *((unsigned char*)&m_pd); }
    unsigned char&  mref() const { return *(((unsigned char*)&m_pd)+1); }
    unsigned short& yref() const { return *(((unsigned short*)&m_pd)+1); }
    unsigned char&  dref() { return *((unsigned char*)&m_pd); }
    unsigned char&  mref() { return *(((unsigned char*)&m_pd)+1); }
    unsigned short& yref() { return *(((unsigned short*)&m_pd)+1); }

    // private data
    unsigned long m_pd;
};

inline int tdate::compare(const tdate& date) const
{
	int cmp;
	cmp = year()*12 + month() - (date.year()*12 + date.month());
	if (cmp == 0) cmp = mday() - date.mday();
	return cmp;
}


/////////////////////////////////////////////////
// Time functions
// ~~~~~~~~~~~~~~
// Hours 0..23, Minutes 0..59, Seconds 0..59.

class ttime
{
 public:
    // constructors
    ttime();
    ttime(int hour, int min=0, int sec=0);
    ttime(const ttime& t);
    ttime(const time_t tt);
    
    // member functions
    bool   isnull() const { return m_pt == -1; }
    ttime& mktime(int h, int m=0, int s=0);
    ttime& current() { return mktime(-1); }
    int    hour() const { return (m_pt/3600); }
    int    min() const { return ((m_pt/60)%60); }
    int    sec() const { return (m_pt%60); }
    int    compare(const ttime& t) const { return m_pt - t.packed_time(); }
    ttime  str2time(const char* buf);
    char*  time2str(char* buf=(char*)0, int sep='\0');
    long   packed_time() const { return m_pt; }
    
    // operators
    ttime& operator = (const ttime& t) { m_pt = t.packed_time(); return *this; }
    bool   operator ==(const ttime& t) const { return compare(t) == 0; }
    bool   operator > (const ttime& t) const { return compare(t) > 0; }
    bool   operator >=(const ttime& t) const { return compare(t) >= 0; }
    bool   operator < (const ttime& t) const { return compare(t) < 0; }
    bool   operator <=(const ttime& t) const { return compare(t) <= 0; }

 protected:
    // private data
    long m_pt;
};

// END NAMESPCE MCL
};

#endif   // _mcl_datetime_h
