/*
 milix Class Library v1.1
 Double-linked lists

 FILE  : dllist.h
 AUTHOR: Anastasios Milikas (aka 'milix'). e-mail: milix@milix.net

 Copyright (c) 1999-2004 by milix
 Permission is hereby granted, free of charge, to any person obtaining a copy of
 this software and associated documentation files (the "Software"), to deal in
 the Software without restriction, including without limitation the rights to use,
 copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
 Software, and to permit persons to whom the Software is furnished to do so,
 subject to the following conditions:

 The above copyright notice and this permission notice shall be included in all
 copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

 Overview
 ~~~~~~~~
 Doubly linked list definition and implementation.
 DLList is a template class.
 Usage: DLList<int>, DLList<MyClass>, ...
*/

#ifndef _mcl_dllist_h
#define _mcl_dllist_h

// BEGIN NAMESPACE MCL
namespace mcl {

/////////////////////////////////////////////////
// dllist
// ~~~~~~
// Doubly linked list.
//
template <class T>
class dllist
{
 public:
	// Constructor, Destructor
	dllist();
	~dllist();

	// Member functions
	bool  empty() const { return bool(count == 0); }
	long  length() const { return count; }
	long  size() const { return count; }
	void  clear();
	long  append(const T&);
	long  push_back(const T& t) { return append(t); }
	long  insert_after(long, const T&);
	long  insert_before(long, const T&);
	void  remove(long);
	void  swap(long p1, long p2);
	long  owns(const T&) const;
	T&    operator [] (long p) { _index(p); return cur->data; }

	// Emulation of FIFO queue
	bool  push(const T& obj) { return bool(insert_before(0, obj) != -1); }
	T     pop();

 private:
	// Private member functions
	void _index(long);

	struct node_t {
		node_t* bk;
		node_t* fd;
		T       data;
	};

	// Private data members
	node_t* h;
	node_t* cur;
	long    curidx;
	long    count;
};


/////////////////////////////////////////////////
// IMPLEMENTATION
/////////////////////////////////////////////////

template <class T>
dllist<T>::dllist()
{
	h = cur = 0;
	curidx = 0;
	count = 0;
}

template <class T>
dllist<T>::~dllist()
{
	clear();
}

template <class T>
void dllist<T>::clear()
{
	if (!h) return;

	node_t* p = h->bk;
	h->bk = 0;
	h = 0;
	while (p) {
		node_t* nxt = p->bk;
		delete p;
		p = nxt;
	}

	h = cur = 0;
	count = 0;
}

template <class T>
long dllist<T>::append(const T& obj)
{
	node_t* t;

	// allocate a node
	if ((t = new node_t) == 0) return -1;
	t->data = obj;

	// add the node to the list
	if (!h) {
        t->fd = t->bk = h = t;
    } else {
		t->fd = h->fd;
		t->fd->bk = t;
		t->bk = h;
		h->fd = t;
	}

	count++;
	return (count-1);
}

template <class T>
long dllist<T>::insert_before(long p, const T& obj)
{
	node_t* t;

	// if list is empty, just append the object
	if (!h) return append(obj);

	// allocate a node
	if ((t = new node_t) == 0) return -1;
	t->data = obj;

	// add the node to the list
	_index(p);
	t->fd = cur->fd;
	t->bk = cur;
	t->bk->fd = t;
	t->fd->bk = t;
	if (p == 0) h = t;
	count++;

	return p;
}

template <class T>
long dllist<T>::insert_after(long p, const T& obj)
{
	node_t* t;

	// if list is empty, just append the object
	if (!h) return append(obj);

	// allocate a node
	if ((t = new node_t) == 0) return -1;
	t->data = obj;

	// add the node to the list
	_index(p);
	t->fd = cur;
	t->bk = cur->bk;
	t->fd->bk = t;
	t->bk->fd = t;
	count++;
	cur = 0;

	return p;
}

template <class T>
void dllist<T>::remove(long p)
{
	node_t* t;

	// find the node
	_index(p);
	t = cur;

	// delete the node
	t->fd->bk = t->bk;
	t->bk->fd = t->fd;
	if (p == 0) h = h->bk;
	delete t;
	if (count == 1) h = 0;
	count--;
	cur = 0;
}

template <class T>
long dllist<T>::owns(const T& obj) const
{
	long p = 0;
	node_t* t = h;
	if (t) {
		do {
			if (t->data == obj) return p;
			t = t->bk;
			p++;
		} while (t != h);
	}
	return -1;
}

template <class T>
T dllist<T>::pop()
{
	T ret = (*this)[0];
	if (count) remove(0);
	return ret;
}

template <class T>
void dllist<T>::swap(long p1, long p2)
{
    T tmp;
	node_t *t1, *t2;
    _index(p1); t1 = cur;
    _index(p2); t2 = cur;
    tmp = t1->data;
    t1->data = t2->data;
    t2->data = tmp;
}

template <class T>
void dllist<T>::_index(long p)
{
	register long i;
	if (cur && p == curidx) return;
	else if (cur && p == curidx+1) cur = cur->bk;
	else if (cur && p == curidx-1) cur = cur->fd;
	else if (p < (count >> 1))
		for (cur = h, i = p ; i > 0 ; i--) cur = cur->bk;
	else for (cur = h->fd, i = (count-p-1) ; i > 0 ; i--) cur = cur->fd;
	curidx = p;
}

// END NAMESPACE MCL
};

#endif   // _mcl_dllist_h
