/****************************************************************************
 bstree.cpp

 milix Class Library
 Author: milix (milix@milix.net)
 Date  : 1999-2004

 Description
 ~~~~~~~~~~~
 Binary search tree class implementation
*****************************************************************************/

#include "../include/bstree.h"
#include "../include/defs.h"
#include <stdlib.h>
#include <string.h>
#include <fcntl.h> 
#if defined(__MSDOS__) || defined(_WIN32)
#include <io.h>
#else
#include <unistd.h>
#endif
#ifdef __GO32__
#include <unistd.h>
#endif

using namespace mcl;

// header
// APPHDR_SZ bytes are reserved for application header. Used in filing
// 20 bytes are used by bstree header
#define APPHDR_SZ   64
#define HDR_SZ      20

// macros
#ifndef ISEVEN
#define ISEVEN(x)   (!((x) & 1))
#endif
#ifndef EVEN
#define EVEN(x)     (((x)+1) & ~1)
#endif
#ifndef MIN
#define MIN(a,b) ((a) < (b) ? (a) : (b))
#endif
#ifndef MAX
#define MAX(a,b) ((a) > (b) ? (a) : (b))
#endif

/////////////////////////////////////////////////
// bstree
// ~~~~~~
// Implementation
//
// NOTES
// ~~~~~
// Tree's header size: 20 bytes (five 32-bit integers)
//    (root)(del)(len)(tmp1)(tmp2)[nodes...]
// If memory pool is OnDisk a file handle must be attached
// before any tree operation (init, insert, remove, find etc.)

bstree::bstree()
{
    m_datasz   = 2;
    m_amount   = 32;
	m_capacity = 0;
	m_count    = 0;
	m_mpool    = OnMem;
	m_space    = NULL;
	m_buf      = NULL;
	reserve(m_amount);
}

bstree::bstree(unsigned datasz, unsigned cap, int mpool, int fh)
{
    m_datasz   = datasz;
    m_amount   = 32;
	m_capacity = 0;
	m_count    = 0;
	m_mpool    = OnMem;
	m_space    = NULL;
	m_buf      = NULL;
    if (mpool == OnDisk) m_space = (void*)fh;
    if (cap > 0) reserve(cap);
	else reserve(m_amount);
}

bstree::~bstree()
{
	if (m_buf) delete m_buf;
	if (m_space && m_mpool == OnMem) delete (char*)m_space;
}

int bstree::clear()
{
	m_count = 0;
	set_root(0);
	set_first_erased(0);
	return 0;
}

int bstree::reserve(unsigned cap)
{
    unsigned sz, savsz;
    unsigned savcap = m_capacity;
    void* saved = m_space;

    if (!m_buf) {
		m_buf = new char [EVEN(MAX(m_datasz, sizeof(node_t)))];
		if (!m_buf) return BAD_ALLOC;
    }

    if (cap <= m_capacity) return SUCCESS;

	if (m_mpool == OnMem) {
        savsz = estim_space_size();
        m_capacity = cap;
        sz = estim_space_size();
		m_space = (void*) new char [HDR_SZ+sz];
		if (!m_space) {
            m_space = saved;
            m_capacity = savcap;
            return BAD_ALLOC;
        }
        if (saved) {
            memcpy(m_space, saved, HDR_SZ+savsz);
            delete (char*)saved;
        }
    } else {
		if ((int)m_space==-1) return BAD_ALLOC;
        lseek((int)m_space, APPHDR_SZ+8L, 0);
		if (::read((int)m_space, &m_count, sizeof(m_count)) <= 0) {
            if (creat((int)m_space) != SUCCESS) return BAD_ALLOC;
        }
    }
    
	if (!m_count) {
		set_root(0);
		set_first_erased(0);
	}

    return SUCCESS;
}

node_t bstree::insert(const void* item)
{
	int cmp;
	node_t n;

	n = new_node();
	if (!n) return (node_t)0;
	set_left(n, 0);
	set_right(n, 0);
	set_data(n, item);

	if (!m_count) {
		set_root(n);
		m_count++;
		return n;
	}

	node_t r, t = root();
	for (;;) {
		r = t;
		cmp = compare(item, data(t));
		if (cmp <= 0) t = left(t);
		else t = right(t);
		if (!t) {
			if (cmp <= 0) set_left(r, n);
			else set_right(r, n);
			m_count++;
			return n;
		}
	}
	return (node_t)0;
}

int bstree::erase(node_t node, node_t parent)
{
	int islt, isrt;
	node_t lt, rt;

	if (!m_count) return SUCCESS;
	if (m_count==1) {
		m_count = 0;
		set_root(0);
		set_first_erased(0);
		return SUCCESS;
	}

	islt = parent && left(parent)==node;
	isrt = parent && right(parent)==node;
	lt = left(node);
	rt = right(node);
	if (lt && !rt) {
		if (islt && parent) set_left(parent, lt);
		else if (isrt && parent) set_right(parent, lt);
		if (!parent) set_root(lt ? lt:rt);
	} else if (rt && !lt) {
		if (islt && parent) set_left(parent, rt);
		else if (isrt && parent) set_right(parent, rt);
		if (!parent) set_root(lt ? lt:rt);
	} else if (lt && rt) {
		if (islt && parent) {
			set_left(parent, rt);
			set_left(left_most(rt), lt);
		} else if (isrt && parent) {
			set_right(parent, lt);
			set_right(right_most(lt), rt);
		} else if (!parent) {
			set_root(lt);
			set_right(right_most(lt), rt);
		}
	} else if (parent) {
		if (islt) set_left(parent, 0);
		else if (isrt) set_right(parent, 0);
	}
	del_node(node);
	m_count--;
	return SUCCESS;
}

node_t bstree::find_first(const void* key, node_t* parent)
{
	int cmp;
	node_t n;

	if (parent) *parent = (node_t)NULL;
	if (m_count) for (n = root(); n;) {
		cmp = compare(key, data(n));
		if (cmp == 0) return n;
		if (parent) *parent = n;
		if (cmp < 0) n = left(n);
		else n = right(n);
	}
	return (node_t)NULL;
}

node_t bstree::find_next(node_t node, const void* key)
{
	int cmp;
	node = left(node);
	cmp = compare(key , data(node));
	if (cmp == 0) return node;
	return (node_t)NULL;
}

void* bstree::data(node_t node)
{
	switch (m_mpool) {
	case OnMem: return (void*)((char*)m_space+node+(sizeof(node_t)<<1));
	case OnDisk:
		if ((int)m_space==-1) break;
		lseek((int)m_space , (long)node+(sizeof(node_t)<<1) , 0);
		if (::read((int)m_space, m_buf, m_datasz) <= 0) break;
		return (void*)m_buf;
	}
	return NULL;
}

void bstree::set_data(node_t node, const void* data)
{
	switch (m_mpool) {
	case OnMem:
		memcpy((char*)m_space+node+(sizeof(node_t)<<1), data, m_datasz);
		break;
	case OnDisk:
		if ((int)m_space==-1) break;
		lseek((int)m_space, (long)node+(sizeof(node_t)<<1) , 0);
		::write((int)m_space, data, m_datasz);
		break;
	}
}

node_t bstree::left(node_t node)
{
	switch (m_mpool) {
	case OnMem: return (node_t) *((node_t*)((char*)m_space+lref(node)));
	case OnDisk:
		if ((int)m_space==-1) break;
		lseek((int)m_space, (long)lref(node), 0);
		if (::read((int)m_space, m_buf, sizeof(node_t)) <= 0) break;
		return (node_t)*((node_t*)m_buf);
	}
	return (node_t)NULL;
}

node_t bstree::right(node_t node)
{
	switch (m_mpool) {
	case OnMem: return (node_t) *((node_t*)((char*)m_space+rref(node)));
	case OnDisk:
		if ((int)m_space==-1) break;
		lseek((int)m_space, (long)rref(node), 0);
		if (::read((int)m_space, m_buf, sizeof(node_t)) <= 0) break;
		return (node_t)*((node_t*)m_buf);
	}
	return (node_t)NULL;
}

void bstree::set_left(node_t node, node_t point)
{
	switch (m_mpool) {
	case OnMem:
		*((node_t*)((char*)m_space+lref(node))) = point;
		break;
	case OnDisk:
		if ((int)m_space==-1) break;
		lseek((int)m_space, (long)lref(node) , 0);
		::write((int)m_space, &point, sizeof(node_t));
		break;
	}
}

void bstree::set_right(node_t node, node_t point)
{
	switch (m_mpool) {
	case OnMem:
		*((node_t*)((char*)m_space+rref(node))) = point;
		break;
	case OnDisk:
		if ((int)m_space==-1) break;
		lseek((int)m_space, (long)rref(node), 0);
		::write((int)m_space, &point, sizeof(node_t));
		break;
	}
}

// linear-time, zero m_m_space overhead binary tree rebalancing from
// Stout & Warren, ``Tree rebalancing in linear space and time''
// CACM, Sept, 1986, p902.
void bstree::balance()
{
	if (m_count <= 2) return;

	node_t r = root();

	// make re-attaching the root easy via trickery
	node_t pseudo_root = (node_t)12;
	set_right(pseudo_root, r);

	// phase 1: tree-to-vine
	node_t vine_tail = pseudo_root;
	node_t remainder = r;

	while (remainder) {
		if (!left(remainder)) {
			vine_tail = remainder;
			remainder = right(remainder);
		} else {
			node_t tmp = left(remainder);
			set_left(remainder, right(tmp));
			set_right(tmp, remainder);
			set_right(vine_tail, remainder=tmp);
		}
	}

	// phase 2: vine-to-tree
	long spines = m_count-1;
	while (spines > 1) {
		long compressions = spines >> 1; // compress every other node
		spines -= compressions + 1;      // halve for next time

		node_t scanner = pseudo_root;
		while (compressions-- > 0) {
			node_t child = right(scanner);
			node_t grandchild = right(child);
			set_right(scanner, grandchild);
			set_right(child, left(grandchild));
			set_left(grandchild, child);
			scanner = grandchild;
		}
	}

	set_root(right(pseudo_root));
}

node_t bstree::root() const
{
	node_t r;
	switch (m_mpool) {
	case OnMem: return *((node_t*)m_space);
	case OnDisk:
		if ((int)m_space==-1) break;
		lseek((int)m_space, APPHDR_SZ, 0);
		if (::read((int)m_space, &r, sizeof(node_t)) <= 0) break;
		return r;
	}
	return (node_t)NULL;
}

void bstree::set_root(node_t root)
{
	switch (m_mpool) {
	case OnMem: *((node_t*)m_space) = root; break;
	case OnDisk:
		if ((int)m_space==-1) break;
		lseek((int)m_space, APPHDR_SZ, 0);
		::write((int)m_space, &root, sizeof(node_t));
    	break;
	}
}

node_t bstree::first_erased() const
{
	node_t firstdel;
	switch (m_mpool) {
	case OnMem: return *(((node_t*)m_space)+1);
	case OnDisk:
		if ((int)m_space==-1) break;
		lseek((int)m_space, APPHDR_SZ+4L, 0);
		if (::read((int)m_space, &firstdel, sizeof(node_t)) <= 0) break;
		return firstdel;
	}
	return (node_t)NULL;
}

void bstree::set_first_erased(node_t firstdel)
{
	switch (m_mpool) {
	case OnMem: *(((node_t*)m_space)+1) = firstdel; break;
	case OnDisk:
		if ((int)m_space==-1) break;
		lseek((int)m_space, APPHDR_SZ+4L, 0);
		::write((int)m_space, &firstdel, sizeof(node_t));
		break;
	}
}

unsigned bstree::estim_space_size()
{
	return m_capacity*(2*sizeof(node_t)+EVEN(m_datasz));
}

node_t bstree::new_node()
{
	node_t node;
	node_t fstdel;
	if (m_count >= m_capacity) {
        if (reserve(m_count+m_amount) != SUCCESS) return (node_t)NULL;
    }
	fstdel = first_erased();
	if (fstdel) {
		node = fstdel;
		set_first_erased(left(fstdel));
	} else {
    	node = HDR_SZ+m_count*(m_datasz+(sizeof(node_t)<<1));
		if (m_mpool != OnMem) node += APPHDR_SZ;
    }
	return node;
}

void bstree::del_node(node_t node)
{
	node_t firstdel;
	if (!m_count) return;
	firstdel = first_erased();
	// add this node to the list of deleted nodes
	if (!firstdel) {
        set_first_erased(node);
    } else {
		register node_t n;
		for (n = firstdel;;) {
			node_t lt = left(n);
			if (lt) n = lt; else break;
		}
		set_left(n, node);
	}
	set_left(node, 0);
	set_right(node, 0);
}

node_t bstree::left_most(node_t node)
{
	register node_t n;
	for (n = node;;) {
		node_t lt = left(n);
		if (lt) n = lt; else break;
	}
	return n;
}

node_t bstree::right_most(node_t node)
{
	register node_t n;
	for (n = node;;) {
		node_t rt = right(n);
		if (rt) n = rt; else break;
	}
	return n;
}

int bstree::creat(int fh)
{
	char hdr[HDR_SZ];
	if (fh==-1 && m_mpool==OnDisk) fh = (int)m_space;
	if (fh==-1) return BAD_OPEN;
	memset(hdr, 0, HDR_SZ);
	lseek(fh, APPHDR_SZ, 0);
	if (::write(fh , hdr , HDR_SZ) <= 0) return BAD_RW;
	return SUCCESS;
}

int bstree::read(int fh)
{
	register unsigned i;
	register char* p = (char*)m_space+HDR_SZ;

	if (fh==-1 && m_mpool==OnDisk && (int)m_space!=-1) return SUCCESS;
	if (fh==-1) return BAD_OPEN;
	lseek(fh, APPHDR_SZ, 0);
	if (m_mpool == OnMem) {
		// read the header
		reserve(m_amount);
		if (::read(fh, m_space, HDR_SZ) <= 0) return BAD_RW;
		m_count = *((node_t*)m_space+2);
		reserve(m_count);
		// read the nodes m_m_space
		for (i = 0 ; i < m_count ; i++) {
			if (::read(fh, p, sizeof(node_t)) <= 0) return BAD_RW;
			p += sizeof(node_t);
			if (::read(fh, p, sizeof(node_t)) <= 0) return BAD_RW;
			p += sizeof(node_t);
			if (::read(fh, p, m_datasz) <= 0) return BAD_RW;
			p += m_datasz;
		}
	}
	
	return SUCCESS;
}

int bstree::write(int fh)
{
	register unsigned i;
	register char* p = (char*)m_space+HDR_SZ;

	// if nodes stored on disk, write only the count
	if (fh==-1 && m_mpool==OnDisk && (int)m_space!=-1) {
		fh = (int)m_space;
		lseek(fh, APPHDR_SZ+8L, 0);
		if (::write(fh, &m_count, sizeof(m_count)) <= 0) return BAD_RW;
		return SUCCESS;
	}

	if (fh == -1) return BAD_OPEN;
	lseek(fh, APPHDR_SZ, 0);
	if (m_mpool == OnMem) {
		// write the header
		if (::write(fh, m_space, HDR_SZ) <= 0) return BAD_RW;
		// write the nodes workspace
		for (i = 0 ; i < m_capacity ; i++) {
    		if (::write(fh, p, sizeof(node_t)) <= 0) return BAD_RW;
			p += sizeof(node_t);
			if (::write(fh, p, sizeof(node_t)) <= 0) return BAD_RW;
			p += sizeof(node_t);
			if (::write(fh, p, sizeof(node_t)) <= 0) return BAD_RW;
			p += m_datasz;
		}
	}
	return SUCCESS;
}

/////////////////////////////////////////////////
// bstreeTraveller
// ~~~~~~~~~~~~~~~
// Travelling the nodes of a binary search tree.
//

bstraveller::bstraveller(bstree* tree)
{
	m_count  = tree->length();
	m_datasz = tree->data_size();
	m_space  = tree->workspace();
	m_mpool  = tree->memory();
	m_buf    = tree->buffer();
}

int bstraveller::travel(int way)
{
	if (!m_count) return BAD_ALLOC;
	return _travel_nodes(root(), way);
}

int bstraveller::compare(const void*, const void*) { return 0; }

int bstraveller::visitor(void*) { return SUCCESS; }

#define CATCH(m) if ((ec = (m)) != SUCCESS) return (ec);

int bstraveller::_travel_nodes(node_t node, int way)
{
	int ec;

	if (node) {
		switch (way) {
		case TravelLeft:
			CATCH(_travel_nodes(left(node) , way))
			CATCH(visitor(data(node)))
			CATCH(_travel_nodes(right(node) , way))
			break;

		case TravelRight:
			CATCH(_travel_nodes(right(node) , way))
			CATCH(visitor(data(node)))
			CATCH(_travel_nodes(left(node) , way))
			break;

		case TravelCurrent:
			CATCH(visitor(data(node)))
			CATCH(_travel_nodes(left(node) , way))
			CATCH(_travel_nodes(right(node) , way))
			break;
		}
	}

	return SUCCESS;
}
