/****************************************************************************
 btree.cpp

 milix Class Library
 Author: milix (milix@milix.net)
 Date  : 1999-2003

 Description
 ~~~~~~~~~~~
 Binary tree class implementation
*****************************************************************************/

#include "../include/btree.h"
#include "../include/strings.h"
#include <fcntl.h>
#if defined(__MSDOS__) || defined(__WIN32__)
#include <io.h>
#else
#include <unistd.h>
#define tell(fh) lseek(fh,0,SEEK_CUR)
#endif
#ifdef __GO32__
#include <unistd.h>
#endif

using namespace mcl;

// btree::node_t ////////////////////////////////
/////////////////////////////////////////////////

unsigned btree::node_t::lnodes()
{
	unsigned nodes;
	register node_t* node;
	for (node = this, nodes = 0 ; node->left ; nodes++, node = node->left);
	return nodes;
}

unsigned btree::node_t::rnodes()
{
	unsigned nodes;
	register node_t* node;
	for (node = this, nodes = 0 ; node->right ; nodes++, node = node->right);
	return nodes;
}

btree::node_t* btree::node_t::left_most()
{
	register node_t* n;
	for (n = this ; n->left ; n = n->left);
	return n;
}

btree::node_t* btree::node_t::right_most()
{
	register node_t* n;
	for (n = this ; n->right ; n = n->right);
	return n;
}

btree::node_t* btree::node_t::left_most_parent()
{
	register node_t* n;
	for (n = this ; n->islt() ; n = n->parent);
	return n;
}

btree::node_t* btree::node_t::right_most_parent()
{
	register node_t* n;
	for (n = this ; n->isrt() ; n = n->parent);
	return n;
}

unsigned btree::node_t::parent_distance(node_t* node)
{
	unsigned dist;
	register node_t* n;

	if (node == this) return 0;
	for (n = this, dist = 0 ; n != node && n ; n = n->parent, dist++);
	if (!n) return 0;
	return dist;
}

unsigned btree::node_t::left_distance(node_t* node)
{
	unsigned dist;
	register node_t* n;

	if (node == this) return 0;
	for (n = this, dist = 0 ; n != node && n ; n = n->left, dist++);
	if (!n) return 0;
	return dist;
}

unsigned btree::node_t::right_distance(node_t* node)
{
	unsigned dist;
	register node_t* n;

	if (node == this) return 0;
	for (n = this, dist = 0 ; n != node && n ; n = n->right, dist++);
	if (!n) return 0;
	return dist;
}

// btree ////////////////////////////////////////
/////////////////////////////////////////////////

btree::btree()
{
	root  = NULL;
	cur   = NULL;
	count = 0;
	sz    = 0;
}

btree::btree(const btree* src)
{
	copy(src);
}

void btree::destroy()
{
	kill(root);
}

void btree::kill(node_t* node)
{
	if (node) {
		kill(node->left);
		kill(node->right);
		release(node->data);
		delete node;
		count--;
	}
}

void btree::clear()
{
	kill(root);
	root  = NULL;
	cur   = NULL;
	count = 0;
}

void btree::copy(const btree* src)
{
	if (root) clear();
	sz   = ((btree*)src)->sizeOf();
	root = copy_nodes(((btree*)src)->getroot());
	cur  = root;
}

inline bool btree::walkleft()
{
	if (!cur || !cur->left) return false;
	cur = cur->left;
	return TRUE;
}

inline bool btree::walkright()
{
	if (!cur || !cur->right) return false;
	cur = cur->right;
	return TRUE;
}

inline bool btree::walkback()
{
	if (!cur || !cur->parent) return false;
	cur = cur->parent;
	return TRUE;
}

inline void btree::set_node_data(node_t* node, const void* data)
{
	memcpy(node->data, data, sz);
}

int btree::addleft(void* item)
{
	node_t* node;

	// prepare the new node
	node = new node_t;
	if (!node) return BAD_ALLOC;
	node->data = new char [sz];
	if (!node->data) return BAD_ALLOC;
	set_node_data(node, item);
	node->left = NULL;
	node->right = NULL;

	// check if tree is empty
	if (!root) {
		root = node;
		cur  = node;
		root->parent = NULL;
		root->depth  = 0;
	} else {
		// walk left the current node and add the new node
		while (walkleft());
		cur->left = node;
		node->parent = cur;
		node->depth = cur->depth;
		cur = node;
	}

	count++;
	return SUCCESS;
}

int btree::addright(void* item)
{
	node_t* node;

	// prepare the new node
	node = new node_t;
	if (!node) return BAD_ALLOC;
	node->data = new char [sz];
	if (!node->data) return BAD_ALLOC;
	set_node_data(node , item);
	node->depth = !cur ? 0 : cur->depth + 1;
	node->left = NULL;
	node->right = NULL;

	// check if tree is empty
	if (!root) {
		root = node;
		cur  = node;
		root->parent = NULL;
	} else {
		// walk right the current node and add the new node
		while (walkright()) node->depth++;
		cur->right = node;
		node->parent = cur;
		cur = node;
	}

	count++;
	return SUCCESS;
}

inline void btree::deleteleft()
{
	if (cur) {
		kill(cur->left);
		cur->left = NULL;
	}
}

inline void btree::deleteright()
{
	if (cur) {
		kill(cur->right);
		cur->right = NULL;
	}
}

bool btree::findline(void* line, unsigned n)
{
	unsigned   i;
	node_t* pnode;
	node_t* node = root;
	char* linep = (char*) line;

	for (i = 0 ; i < n ; i++) {
tryAgain:
		if (!node) return false;
		if (!memcmp(node->data, linep, sz)) {
			pnode = node;
			node  = node->right;
			linep += sz;
		} else {
			node = node->left;
			goto tryAgain;
		}
	}

	cur = pnode;
	return true;
}

int btree::insertline(void* line, unsigned n)
{
	unsigned   i;
	node_t* pnode;
	node_t* node = root;
	bool  newline = FALSE;
	char* linep = (char*) line;

	// find the position to append the new line
	for (i = 0 ; i < n ; i++) {
tryAgain:
		if (!node) break;
		if (!memcmp(node->data, linep, sz)) {
			pnode   = node;
			node    = node->right;
			newline = false;
			linep += sz;
		} else {
			pnode   = node;
			node    = node->left;
			newline = true;
			goto tryAgain;
		}
	}

	// the line already exists
	if (i == n) return SUCCESS;

	// place the new elements of the line to the tree
	for (; i < n ; i++) {
		node = new node_t;
		if (node == NULL) return BAD_ALLOC;
		node->data  = new char [sz];
		if (!node->data) return BAD_ALLOC;
		set_node_data(node , linep);
		node->left  = NULL;
		node->right = NULL;
		if (!root) {
			root         = node;
			pnode        = node;
			root->depth  = 0;
			root->parent = NULL;
		} else {
			node->parent = pnode;
			if (newline) {
				pnode->left  = node;
				node->depth  = pnode->depth;
			} else {
				pnode->right = node;
				node->depth  = pnode->depth + 1;
			}
			pnode = node;
			newline = false;
		}
		linep += sz;
		count++;
		cur = node;
	}

	return SUCCESS;
}

int btree::deleteline(void* /*line*/, unsigned /*n*/)
{
	return SUCCESS;
}

int btree::promoteline(void* /*line*/, unsigned /*n*/)
{
	return SUCCESS;
}

int btree::demoteline(void* /*line*/, unsigned /*n*/)
{
	return SUCCESS;
}

#define HAS_LEFT  0x01
#define HAS_RIGHT 0x10

int btree::read(int fh)
{
	if (fh == -1) return BAD_OPEN;
	if (root) clear();

	// read number of items
	if (::read(fh, &count , sizeof(unsigned)) != sizeof(unsigned)) return BAD_RW;
	if (count == 0) return BAD_NUMBER;

	// read nodes
	root = read_nodes(fh);

	return (!root) ? BAD_RW : SUCCESS;
}

int btree::write(int fh)
{
	if (fh == -1) return BAD_OPEN;
	if (!root) return BAD_ALLOC;

	// write number of items
	if (::write(fh, &count, sizeof(unsigned)) != sizeof(unsigned)) return BAD_RW;

	// write nodes
	return write_nodes(root , fh);
}

int btree::read_node_data(int fh, void* data)
{
	if (::read(fh, data, sz) != sz) return BAD_RW;
	return SUCCESS;
}

int btree::write_node_data(int fh, void* data)
{
	if (::write(fh, data, sz) != sz) return BAD_RW;
	return SUCCESS;
}

btree::node_t* btree::read_nodes(int fh)
{
	unsigned char flags;
	node_t* node;

	node = new node_t;
	if (!node) return NULL;
	node->data = new char [sz];
	if (!node->data) return NULL;
	if (read_node_data(fh , node->data) != SUCCESS) return NULL;
	if (::read(fh , &node->depth , sizeof(unsigned)) != sizeof(unsigned)) return NULL;
	node->left   = NULL;
	node->right  = NULL;
	node->parent = NULL;

	if (::read(fh , &flags , 1) != 1) return NULL;
	if (flags & HAS_LEFT) {
		node->left = read_nodes(fh);
		if (node->left) node->left->parent = node;
		else return NULL;
	}
	if (flags & HAS_RIGHT) {
		node->right = read_nodes(fh);
		if (node->right) node->right->parent = node;
		else return NULL;
	}

	return node;
}

int btree::write_nodes(node_t* node, int fh)
{
	unsigned char flags = 0;

	if (!node) return SUCCESS;

tailRecursion:
	if (write_node_data(fh , node->data) != SUCCESS) return BAD_RW;
	if (::write(fh, &node->depth, sizeof(unsigned)) != sizeof(unsigned)) return BAD_RW;
	if (node->left) flags |= HAS_LEFT;
	if (node->right) flags |= HAS_RIGHT;
	if (::write(fh , &flags , 1) != 1) return BAD_RW;

	// recursion reduction for left nodes
	if (node->left && !node->right) {
		flags = 0;
		node  = node->left;
		goto tailRecursion;
	} else if (write_nodes(node->left, fh) != SUCCESS) return BAD_RW;

	// recursion reduction for right nodes
	if (node->right && !node->left) {
		flags = 0;
		node  = node->right;
		goto tailRecursion;
	} else if (write_nodes(node->right, fh) != SUCCESS) return BAD_RW;

	return SUCCESS;
}

btree::node_t* btree::copy_nodes(node_t* src)
{
	if (!src) {
        return NULL;
    } else {
		node_t* node = new node_t;
		if (!node) return NULL;
		node->data = new char [sz];
		if (!node->data) return NULL;
		set_node_data(node , src->data);
		node->depth = src->depth;
		node->left = copy_nodes(src->left);
		node->right = copy_nodes(src->right);
		if (node->left) node->left->parent = node;
		if (node->right) node->right->parent = node;
		return node;
	}
}

// btraveller ///////////////////////////////
/////////////////////////////////////////////////

btraveller::btraveller(btree* tree)
{
	root  = tree->getroot();
	cur   = tree->current();
	count = tree->length();
	sz    = tree->sizeOf();
}
