/****************************************************************************
 datetime.cpp

 milix Class Library
 Author: milix (milix@milix.net)
 Date  : 1999-2003

 Description
 ~~~~~~~~~~~
 Date/time functions implementation
*****************************************************************************/

#include "../include/datetime.h"
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

using namespace mcl;

static const int _MON_DAYS[2][12] = {
	{ 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
	{ 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 }
};

/////////////////////////////////////////////////
// tdate
// Month 1..12, Day 1..31.
// The first legal date is 01-01-1900

tdate::tdate()
{
    m_pd = 0;
}

tdate::tdate(int year, int mon, int mday)
{
    mkdate(year, mon, mday);
}

tdate::tdate(const tdate& date)
{
    m_pd = date.packed_date();
}

tdate::tdate(const time_t tt)
{
	struct tm* tm;
	tm = localtime((time_t*)&tt);
	mkdate(tm->tm_year+1900, tm->tm_mon+1, tm->tm_mday);
}

tdate& tdate::mkdate(int year, int mon, int mday)
{
	time_t tt;
	struct tm* tm;
	
	if (year<1900 || mon<1 || mon>12 ||
		mday<1 || mday>_MON_DAYS[isleap()][mon-1]) {
	    tt = time((time_t*)0);
        tm = localtime(&tt);
        year = tm->tm_year+1900;
        mon = tm->tm_mon+1;
        mday = tm->tm_mday;
    }
    dref() = mday;
	mref() = mon;
	yref() = year;
	return *this;
}

int tdate::wday() const
{
	int a, y, m, d;
	a = (14 - mref()) / 12;
	y = yref() - a;
	m = mref() + 12*a - 2;
	d = (dref() + y + y/4 - y/100 + y/400 + (31*m)/12) % 7;
	return d;
}

int tdate::yday() const
{
	time_t tt;
	struct tm* tm;
	tt = time(NULL);
	tm = localtime(&tt);
	tm->tm_year = year() - 1900;
	tm->tm_mon  = month() - 1;
	tm->tm_mday = mday();
	tt = ::mktime(tm);
	tm = localtime((time_t*)&tt);
	return tm->tm_yday;
}

inline int tdate::mdays() const
{
	return _MON_DAYS[isleap()][mref()-1];
}

tdate& tdate::nextday()
{
	dref()++;
	if (dref() > _MON_DAYS[isleap()][mref()-1]) {
		dref() = 1;
		mref()++;
		if (mref() == 13) {
			mref() = 1;
			yref()++;
		}
	}
	return *this;
}

tdate& tdate::prevday()
{
	dref()--;
	if (dref() < 1) {
		mref()--;
		if (mref() < 1) {
			mref() = 12;
			yref()--;
			if (yref() < 1900) {
                m_pd = 0;
                return *this;
            }
		}
		dref() = _MON_DAYS[isleap()][mref()-1];
	}
	return *this;
}

tdate& tdate::nextmon()
{
	mref()++;
	if (mref() > 12) {
		mref() = 1;
		yref()++;
	}
	return *this;
}

tdate& tdate::prevmon()
{
	mref()--;
	if (mref() < 1) {
		mref() = 12;
		yref()--;
		if (yref() < 1900) {
            m_pd = 0;
            return *this;
        }
	}
	return *this;
}

tdate& tdate::add(int days)
{
	if (days > 0) {
        while (days > 0 && !isnull()) {
            days--;
            nextday();
        }
	} else {
        while (days < 0 && !isnull()) {
            days++;
            prevday();
        }
	}
	return *this;
}

int tdate::diff(const tdate& date)
{
	int diff;
	diff = year()*365 - date.year()*365;
	diff += yday() - date.yday();
	return diff;
}

tdate tdate::easter_greece()
{
    tdate e;
	int g, i, j, l, d, m;

	if (year() < 1900) return e;
	g = year() % 19;
	i = (19*g+15) % 30;
	j = (year()+year()/4+i) % 7;
	l = i - j;
	m = 3 + (l+40)/44;
	d = l + 28 - 31*(m/4) + ((year()<2100) ? 13:14);
	if (d > 30) { d -= 30; m++; }
    e.mkdate(year(), m, d);
	return e;
}

char* tdate::date2str(char* buf, int sep)
{
	static char s[16];
	char* p = buf ? buf : s;
	if (sep == '\0') sprintf(p , "%04d%02d%02d", yref(), mref(), dref());
	else sprintf(p , "%04d%c%02d%c%02d", yref(), sep, mref(), sep, dref());
	return p;
}

// str2date recognizes the following formats: yyyymmdd, yyyy-mm-dd, yyyy/mm/dd
tdate tdate::str2date(const char* buf)
{
	tdate e;
	char s[16];
	int y, m, d;

	strncpy(s, buf, 4); s[0] = 0;
	y = atoi(s);
	buf += (buf[4]=='-'||buf[4]=='/'||buf[4]=='.') ? 5 : 4;
	s[0] = buf[0]; s[1] = buf[1]; s[2] = '\0';
	m = atoi(s);
	buf += (buf[2]=='-'||buf[2]=='/'||buf[2]=='.') ? 3 : 2;
	d = atoi(buf);
	e.mkdate(y, m, d);
	return e;
}


/////////////////////////////////////////////////
// Time functions
// ~~~~~~~~~~~~~~
// Hours 0..23, Minutes 0..59, Seconds 0..59.

ttime::ttime()
{
    m_pt = -1;
}

ttime::ttime(int h, int m, int s)
{
    mktime(h, m, s);
}

ttime::ttime(const ttime& t)
{
    m_pt = t.packed_time();
}

ttime::ttime(const time_t tt)
{
    struct tm* tm;
	tm = localtime((time_t*)&tt);
	mktime(tm->tm_hour, tm->tm_min, tm->tm_sec);
}

ttime& ttime::mktime(int h, int m, int s)
{
	time_t tt;
	struct tm* tm;
	
	if (h == -1) {
	   tt = time((time_t*)0);
	   tm = localtime(&tt);
	   h  = tm->tm_hour;
	   m  = tm->tm_min;
	   s  = tm->tm_sec;
    }
    m_pt = h*3600 + m*60 + s;
    return *this;
}

char* ttime::time2str(char* buf, int sep)
{
	static char s[16];
	char* p = buf ? buf : s;
	if (sep == '\0') sprintf(p , "%02d%02d%02d", hour(), min(), sec());
	else sprintf(p , "%02d%c%02d%c%02d", hour(), sep, min(), sep, sec());
	return p;
}

// str2time recognizes the following formats: hhmmss, hh:mm:ss
ttime ttime::str2time(const char* buf)
{
	char t[4];
	int h, m, s;

	t[0] = buf[0]; t[1] = buf[1]; t[2] = '\0';
	h = atoi(t);
	buf += (buf[2]==':') ? 3 : 2;
	t[0] = buf[0]; t[1] = buf[1]; 
	m = atoi(t);
	buf += (buf[2]==':') ? 3 : 2;
	s = atoi(buf);
	return mktime(h, m, s);
}

