/****************************************************************************
 expreval.cpp

 milix Class Library
 Author: milix (milix@milix.net)
 Date  : 1999-2003

 Description
 ~~~~~~~~~~~
 Expresion evaluator class implementation
*****************************************************************************/

#include "../include/expreval.h"
#include <math.h>
#include <stdlib.h>
#include <stdarg.h>

using namespace mcl;

expreval::expreval()
{
	ec = SUCCESS;
}

#define OPERS "+*-/()^|&!~=<>"

int expreval::is_oper(const char* f)
{
	register char c;
	register char* p;

	c = *f;
	if (c >= '0' && c <= '9' || c == '.') return op_none;
	for (p = (char*)OPERS ; *p ; p++) {
		if (c == *p) {
			switch (c) {
				case '(': return op_lp; case ')': return op_rp;
				case '+': return op_add; case '-': return op_sub;
				case '*': return op_mul; case '/': return op_div;
				case '^': return op_pow; case '~': return op_xor;
				case '|': return op_or; case '&': return op_and;
				case '=': return op_eq;
			}
			if (c == '>') return (*(f+1) == '=') ? op_ge : op_gt;
			if (c == '<') return (*(f+1) == '=') ? op_le : op_lt;
			if (c == '!') return (*(f+1) == '=') ? op_ne : op_not;
		}
	}
	return op_none;
}

int expreval::priority(int op)
{
	switch (op) {
		case op_neg: return 1;
		case op_not: return 2;
		case op_pow: return 3;
		case op_mul: case op_div: case op_and: return 4;
		case op_add: case op_sub: case op_or: case op_xor: return 5;
		case op_lt: case op_le: case op_eq: case op_ne: case op_ge: case op_gt:
			return 6;
		case op_lp: return 7;
	}
	return 0;
}

#define STACK_SZ 128
#define TOKEN_SZ 128

int expreval::compile(const char* expr)
{
	int sp, ss;
	int op, sop;
	bool nextneg;
	unsigned nops;
	char token[TOKEN_SZ];
	char stack[STACK_SZ];
	register char *p, *s;

	// initialize
	ec = SUCCESS;
	sp = 0;
	nops = 0;
	tokens = 0;
	nextneg = TRUE;
	compiled = NULL;

	// calculate the stack sizes
	for (p = (char*)expr ; *p ; p++)
		if (is_oper(p) != op_none) nops++;
	if (!nops || nops > STACK_SZ) return ec = BAD_NAME;

	// allocate stacks
	ss = 2*nops+16;
	compiled = new arg_t [ss];
	if (!compiled) return ec = BAD_ALLOC;

	// compile the expesion
	p = (char*)expr;
	for (;;) {
		op = is_oper(p);
		if (op == op_none) {
			if (*p==' ' || *p=='\t' || *p=='\r') { p++; continue; }
			if (*p=='\n' || !*p) break;
			if (*p=='@') {
				compiled[tokens++].what = arg_var;
				p++;
			} else {
				s = token;
				while (*p && *p != ' ' && *p != '\n' && is_oper(p) == op_none)
					*s++ = *p++;
				*s = '\0';
				LINK_VAL(atof(token));
			}
			nextneg = FALSE;
		} else {
			if (op == op_sub && nextneg) {
				op = op_neg;
				nextneg = FALSE;
			}
			if (op == op_lp) {
				stack[sp++] = op_lp;
				nextneg = TRUE;
			} else if (op == op_rp) {
				nextneg = FALSE;
				for (sop = stack[--sp]; sop != op_lp && sp; sop = stack[--sp])
                    LINK_OP((op_t)sop);
			} else if (!sp || sp && priority(op) < priority(stack[sp-1])) {
				stack[sp++] = op;
				nextneg = FALSE;
			} else {
				nextneg = FALSE;
				sop = op;
				while (sp && priority(sop) >= priority(stack[sp-1])) {
					op = stack[--sp];
					LINK_OP((op_t)op);
				}
				op = sop;
				stack[sp++] = sop;
			}
			switch (op) {
				case op_ge: case op_le: case op_ne: p += 2;
				default: p++;
			}
		}
		if (!*p) {
			while (sp) {
				op = stack[--sp];
				LINK_OP((op_t)op);
			}
			break;
		}
		if (sp >= ss) return ec = BAD_INDEX;
	}
	compiled[tokens].what = arg_end;
	return ec;
}

double expreval::calculate(const char* expr, ...)
{
	arg_t* a;
	va_list args;
	double v1, v2;
	register int i, sp = 0;
	double stack[STACK_SZ/2];

	stack[0] = -0.0;
	if (compile(expr) != SUCCESS) goto ret;
	va_start(args , expr);
	for (i = 0 ; i < (int)tokens ; i++) {
		a = &compiled[i];
		if (a->what == arg_end) break;
		if (a->what == arg_val) stack[sp++] = a->arg.val;
		else if (a->what == arg_var) stack[sp++] = va_arg(args , double);
		else {
			if (a->arg.op == op_neg) stack[sp-1] = -stack[sp-1];
			else {
				v1 = stack[--sp];
				v2 = stack[--sp];
				switch (a->arg.op) {
					case op_add: stack[sp++] = v2 + v1; break;
					case op_sub: stack[sp++] = v2 - v1; break;
					case op_mul: stack[sp++] = v2 * v1; break;
					case op_div: stack[sp++] = v2 / v1; break;
					case op_pow: stack[sp++] = pow(v2, v1); break;
					case op_and: stack[sp++] = ((int)v2 && (int)v1); break;
					case op_or:  stack[sp++] = ((int)v2 || (int)v1); break;
					case op_xor: stack[sp++] = ((int)v2 ^ (int)v1); break;
					case op_eq:  stack[sp++] = (v2 == v1); break;
					case op_ne:  stack[sp++] = (v2 != v1); break;
					case op_ge:  stack[sp++] = (v2 >= v1); break;
					case op_le:  stack[sp++] = (v2 <= v1); break;
					case op_gt:  stack[sp++] = (v2 > v1); break;
					case op_lt:  stack[sp++] = (v2 < v1); break;
				}
			}
		}
	}

ret:
	va_end(args);
	if (compiled) delete compiled;
	return stack[0];
}
