/****************************************************************************
 md5.cpp

 milix Class Library
 Author: milix (milix@milix.net)
 Date  : 1999-2003

 Description
 ~~~~~~~~~~~
 MD5 digest algorithm class implementation
*****************************************************************************/

#include "../include/md5.h"
#include <string.h>
#include <stdio.h>
#if defined(_WIN32)
#include <io.h>
#else
#include <unistd.h>
#endif

using namespace mcl;

#define INLINED_MEMCPY(d,s,l) memcpy(d, s, l)
#define INLINED_MEMSET(d,v,l) memset(d, v, l)

md5::md5()
{
	init();
}

md5::md5(int fh)
{
	init();
	update(fh);
	finalize();
}

// MD5 block update operation. Continues an MD5 message-digest
// operation, processing another message block, and updating the
// context.
void md5::update(unsigned char* input, unsigned len)
{
	unsigned inidx, bufidx;
	unsigned bufspace;

	if (m_finalized) return;

	// compute number of unsigned chars mod 64
	bufidx = (unsigned)((m_count[0] >> 3) & 0x3F);

	// update number of bits
	if ((m_count[0] += (len << 3))<(len << 3)) m_count[1]++;

	m_count[1] += (len >> 29);
	bufspace = 64 - bufidx;

	// transform as many times as possible.
	if (len >= bufspace)
	{
		// fill the rest of the buffer and transform
		INLINED_MEMCPY(m_buffer+bufidx , input , bufspace);
		transform(m_buffer);

		// transform each 64-unsigned char piece of the input, bypassing the buffer
		for (inidx = bufspace; inidx+63 < len; inidx += 64) 
			transform(input+inidx);
		bufidx = 0;
	}
	else inidx = 0;

	INLINED_MEMCPY(m_buffer+bufidx , input+inidx, len-inidx);
}

// MD5 update for files.
// Like above, except that it works on files (and uses above as a primitive.)
void md5::update(int fh)
{
/*	unsigned char buf[1024];
	register short len;
	while ((len = read(fh , buf , 1024)) > 0) update(buf , (unsigned)len);
*/
	unsigned char buf[1024];
	int i = 1;
	register short len;
	long flen = 0;
	while ((len = read(fh , buf , 1024)) > 0) 
	{
		flen += (long)len;
		update(buf , (unsigned)len);
		printf("try %d\n", i++);
	}
	printf("bytes: %ld\n", flen);
}

// MD5 finalization. Ends an MD5 message-digest operation, writing the
// the message digest and zeroizing the context.
void md5::finalize()
{
	unsigned char bits[8];
	unsigned idx, padlen;
	static unsigned char PADDING[64] = {
		0x80, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
	};

	if (m_finalized) return;

	// save number of bits
	encode(bits , m_count , 8);

	// pad out to 56 mod 64.
	idx = (m_count[0] >> 3) & 0x3f;
	padlen = (idx < 56) ? (56-idx) : (120-idx);
	update(PADDING , padlen);

	// append length (before padding)
	update(bits , 8);

	// store state in digest
	encode(m_digest , m_state , 16);

	// zeroize sensitive information
	INLINED_MEMSET(m_buffer , 0 , sizeof(*m_buffer));
	m_finalized = TRUE;
}

unsigned char* md5::raw_digest(unsigned char* buf)
{
	if (!m_finalized) return (unsigned char*)0;
	if (!buf) return m_digest;
	INLINED_MEMCPY(buf , m_digest , 16);
	return buf;
}

char* md5::hex_digest(char* buf)
{
	register int i;
	static char hdig[33];

	if (!m_finalized) return (char*)0;
	if (!buf) buf = hdig;
	for (i = 0 ; i < 16 ; i++) sprintf(buf+i*2 , "%02x" , m_digest[i]);
	buf[32] = '\0';
	return buf;
}

void md5::init()
{
	m_finalized = 0;
	m_count[0]  = 0;
	m_count[1]  = 0;
	m_state[0]  = 0x67452301;
	m_state[1]  = 0xefcdab89;
	m_state[2]  = 0x98badcfe;
	m_state[3]  = 0x10325476;
}

// Constants for MD5 transform routine.
// Although we could use C++ style constants, defines are actually better,
// since they let us easily evade scope clashes.
#define S11 7
#define S12 12
#define S13 17
#define S14 22
#define S21 5
#define S22 9
#define S23 14
#define S24 20
#define S31 4
#define S32 11
#define S33 16
#define S34 23
#define S41 6
#define S42 10
#define S43 15
#define S44 21

// MD5 basic transformation. Transforms state based on block.
void md5::transform (unsigned char block[64])
{
	unsigned a = m_state[0], b = m_state[1], c = m_state[2], d = m_state[3], x[16];
	decode(x , block , 64);
	if (m_finalized) return;

	// Round 1
	FF(a, b, c, d, x[ 0], S11, 0xd76aa478); /* 1 */
	FF(d, a, b, c, x[ 1], S12, 0xe8c7b756); /* 2 */
	FF(c, d, a, b, x[ 2], S13, 0x242070db); /* 3 */
	FF(b, c, d, a, x[ 3], S14, 0xc1bdceee); /* 4 */
	FF(a, b, c, d, x[ 4], S11, 0xf57c0faf); /* 5 */
	FF(d, a, b, c, x[ 5], S12, 0x4787c62a); /* 6 */
	FF(c, d, a, b, x[ 6], S13, 0xa8304613); /* 7 */
	FF(b, c, d, a, x[ 7], S14, 0xfd469501); /* 8 */
	FF(a, b, c, d, x[ 8], S11, 0x698098d8); /* 9 */
	FF(d, a, b, c, x[ 9], S12, 0x8b44f7af); /* 10 */
	FF(c, d, a, b, x[10], S13, 0xffff5bb1); /* 11 */
	FF(b, c, d, a, x[11], S14, 0x895cd7be); /* 12 */
	FF(a, b, c, d, x[12], S11, 0x6b901122); /* 13 */
	FF(d, a, b, c, x[13], S12, 0xfd987193); /* 14 */
	FF(c, d, a, b, x[14], S13, 0xa679438e); /* 15 */
	FF(b, c, d, a, x[15], S14, 0x49b40821); /* 16 */

	// Round 2
	GG(a, b, c, d, x[ 1], S21, 0xf61e2562); /* 17 */
	GG(d, a, b, c, x[ 6], S22, 0xc040b340); /* 18 */
	GG(c, d, a, b, x[11], S23, 0x265e5a51); /* 19 */
	GG(b, c, d, a, x[ 0], S24, 0xe9b6c7aa); /* 20 */
	GG(a, b, c, d, x[ 5], S21, 0xd62f105d); /* 21 */
	GG(d, a, b, c, x[10], S22, 0x02441453); /* 22 */
	GG(c, d, a, b, x[15], S23, 0xd8a1e681); /* 23 */
	GG(b, c, d, a, x[ 4], S24, 0xe7d3fbc8); /* 24 */
	GG(a, b, c, d, x[ 9], S21, 0x21e1cde6); /* 25 */
	GG(d, a, b, c, x[14], S22, 0xc33707d6); /* 26 */
	GG(c, d, a, b, x[ 3], S23, 0xf4d50d87); /* 27 */
	GG(b, c, d, a, x[ 8], S24, 0x455a14ed); /* 28 */
	GG(a, b, c, d, x[13], S21, 0xa9e3e905); /* 29 */
	GG(d, a, b, c, x[ 2], S22, 0xfcefa3f8); /* 30 */
	GG(c, d, a, b, x[ 7], S23, 0x676f02d9); /* 31 */
	GG(b, c, d, a, x[12], S24, 0x8d2a4c8a); /* 32 */

	// Round 3
	HH(a, b, c, d, x[ 5], S31, 0xfffa3942); /* 33 */
	HH(d, a, b, c, x[ 8], S32, 0x8771f681); /* 34 */
	HH(c, d, a, b, x[11], S33, 0x6d9d6122); /* 35 */
	HH(b, c, d, a, x[14], S34, 0xfde5380c); /* 36 */
	HH(a, b, c, d, x[ 1], S31, 0xa4beea44); /* 37 */
	HH(d, a, b, c, x[ 4], S32, 0x4bdecfa9); /* 38 */
	HH(c, d, a, b, x[ 7], S33, 0xf6bb4b60); /* 39 */
	HH(b, c, d, a, x[10], S34, 0xbebfbc70); /* 40 */
	HH(a, b, c, d, x[13], S31, 0x289b7ec6); /* 41 */
	HH(d, a, b, c, x[ 0], S32, 0xeaa127fa); /* 42 */
	HH(c, d, a, b, x[ 3], S33, 0xd4ef3085); /* 43 */
	HH(b, c, d, a, x[ 6], S34, 0x04881d05); /* 44 */
	HH(a, b, c, d, x[ 9], S31, 0xd9d4d039); /* 45 */
	HH(d, a, b, c, x[12], S32, 0xe6db99e5); /* 46 */
	HH(c, d, a, b, x[15], S33, 0x1fa27cf8); /* 47 */
	HH(b, c, d, a, x[ 2], S34, 0xc4ac5665); /* 48 */

	// Round 4
	II(a, b, c, d, x[ 0], S41, 0xf4292244); /* 49 */
	II(d, a, b, c, x[ 7], S42, 0x432aff97); /* 50 */
	II(c, d, a, b, x[14], S43, 0xab9423a7); /* 51 */
	II(b, c, d, a, x[ 5], S44, 0xfc93a039); /* 52 */
	II(a, b, c, d, x[12], S41, 0x655b59c3); /* 53 */
	II(d, a, b, c, x[ 3], S42, 0x8f0ccc92); /* 54 */
	II(c, d, a, b, x[10], S43, 0xffeff47d); /* 55 */
	II(b, c, d, a, x[ 1], S44, 0x85845dd1); /* 56 */
	II(a, b, c, d, x[ 8], S41, 0x6fa87e4f); /* 57 */
	II(d, a, b, c, x[15], S42, 0xfe2ce6e0); /* 58 */
	II(c, d, a, b, x[ 6], S43, 0xa3014314); /* 59 */
	II(b, c, d, a, x[13], S44, 0x4e0811a1); /* 60 */
	II(a, b, c, d, x[ 4], S41, 0xf7537e82); /* 61 */
	II(d, a, b, c, x[11], S42, 0xbd3af235); /* 62 */
	II(c, d, a, b, x[ 2], S43, 0x2ad7d2bb); /* 63 */
	II(b, c, d, a, x[ 9], S44, 0xeb86d391); /* 64 */

	m_state[0] += a;
	m_state[1] += b;
	m_state[2] += c;
	m_state[3] += d;

	// zeroize sensitive information.
	INLINED_MEMSET((unsigned char*)x , 0 , sizeof(x));
}

// Encodes input (32-bit) into output (8-bit). Assumes len is
// a multiple of 4.
void md5::encode(unsigned char* output, unsigned* input, unsigned len)
{
	register unsigned i, j;

	for (i = 0, j = 0; j < len; i++, j += 4)
	{
		output[j]   = (unsigned char)(input[i] & 0xff);
		output[j+1] = (unsigned char)((input[i] >> 8) & 0xff);
		output[j+2] = (unsigned char)((input[i] >> 16) & 0xff);
		output[j+3] = (unsigned char)((input[i] >> 24) & 0xff);
	}
}

// Decodes input (unsigned char) into output (UINT4). Assumes len is
// a multiple of 4.
void md5::decode(unsigned* output, unsigned char* input, unsigned len)
{
	register unsigned i, j;

	for (i = 0, j = 0; j < len; i++, j += 4)
		 output[i] = ((unsigned)input[j]) | (((unsigned)input[j+1]) << 8) |
					 (((unsigned)input[j+2]) << 16) | (((unsigned)input[j+3]) << 24);
}

// ROTATE_LEFT rotates x left n bits.
inline unsigned md5::rotate_left(unsigned x, int n)
{
#if defined(__GNUC__) && defined(__i386__)
	__asm__ __volatile__ ("roll %%cl,%0"
		:"=r" (x)
		:"0" (x),"c" (n));
	return x;
#else
	return (x << n) | (x >> (32-n));
#endif
}

// F, G, H and I are basic MD5 functions.
inline unsigned md5::F(unsigned x, unsigned y, unsigned z)
{
	return (x & y) | (~x & z);
}

inline unsigned md5::G(unsigned x, unsigned y, unsigned z)
{
	return (x & z) | (y & ~z);
}

inline unsigned md5::H(unsigned x, unsigned y, unsigned z)
{
	return x ^ y ^ z;
}

inline unsigned md5::I(unsigned x, unsigned y, unsigned z)
{
	return y ^ (x | ~z);
}

// FF, GG, HH, and II transformations for rounds 1, 2, 3, and 4.
// Rotation is separate from addition to prevent recomputation.

inline void md5::FF(unsigned& a, unsigned b, unsigned c, unsigned d,
					unsigned x, unsigned  s, unsigned ac)
{
	a += F(b, c, d) + x + ac;
	a = rotate_left(a, s) + b;
}

inline void md5::GG(unsigned& a, unsigned b, unsigned c, unsigned d,
					unsigned x, unsigned s, unsigned ac)
{
	a += G(b, c, d) + x + ac;
	a = rotate_left(a, s) + b;
}

inline void md5::HH(unsigned& a, unsigned b, unsigned c, unsigned d,
					unsigned x, unsigned s, unsigned ac)
{
	a += H(b, c, d) + x + ac;
	a = rotate_left(a, s) + b;
}

inline void md5::II(unsigned& a, unsigned b, unsigned c, unsigned d,
					unsigned x, unsigned s, unsigned ac)
{
	a += I(b, c, d) + x + ac;
	a = rotate_left (a, s) + b;
}
