/****************************************************************************
 token.cpp

 milix Class Library
 Author: milix (milix@milix.net)
 Date  : 1999-2003

 Description
 ~~~~~~~~~~~
 Text tokenizer class implementation
*****************************************************************************/

#include "../include/defs.h"
#include "../include/token.h"
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#if defined(_WIN32)
#include <io.h>
#else
#include <unistd.h>
#endif

using namespace mcl;

// Lex //////////////////////////////////////////
/////////////////////////////////////////////////

tokenizer::tokenizer()
{
	m_separ    = strdup(" ");
	m_sticky   = strdup("");
	m_rem      = strdup("///**/");
	m_delim    = strdup("\"\"");
	m_incdelim = true;
	m_buff     = NULL;
	m_buffsz   = 0;
	m_ofs      = 0;
	m_fofs     = 0L;
}

tokenizer::~tokenizer()
{
	if (m_separ) delete m_separ;
	if (m_rem) delete m_rem;
	if (m_delim) delete m_delim;
	if (m_sticky) delete m_sticky;
	if (m_buff) delete m_buff;
	_kill_list_entries();
}

int tokenizer::set_separators(const char* s)
{
	if (m_separ) delete m_separ;
	if ((m_separ=strdup(s)) == NULL) return BAD_ALLOC;
	return SUCCESS;
}

int tokenizer::set_sticky_separators(const char* s)
{
	if (m_sticky) delete m_sticky;
	if ((m_sticky = strdup(s)) == NULL) return BAD_ALLOC;
	return SUCCESS;
}

int tokenizer::set_string_delimeters(const char* s, bool include)
{
	if (m_delim) delete m_delim;
	if ((m_delim = strdup(s)) == NULL) return BAD_ALLOC;
	m_incdelim = include;
	return SUCCESS;
}

int tokenizer::set_comment(const char* s)
{
	if (m_rem) delete m_rem;
	if ((m_rem = strdup(s)) == NULL) return BAD_ALLOC;
	return SUCCESS;
}

int tokenizer::set_buffer(const char* s, long bytes)
{
	// if user specifies the buffer size to 0, determine it
	if (bytes == 0) {
		bytes = strlen(s);
		if (bytes == 0) return BAD_NUMBER;
	}

	// release the previous buffer if necessary
	if (m_buff && m_buffsz < bytes) {
		delete m_buff;
		m_buff = NULL;
	}
	if (!m_buff) m_buff = new char [bytes];
	if (!m_buff) return BAD_ALLOC;
	m_buffsz = bytes;
	memcpy(m_buff , s , bytes);

	return SUCCESS;
}

int tokenizer::read_buffer(int fh, long bytes)
{
	register long i;

	if (m_buff && m_buffsz < bytes) {
		delete m_buff;
		m_buff = NULL;
	}
	if (!m_buff) m_buff = new char [bytes+1];
	if (!m_buff) return BAD_ALLOC;
	m_buffsz = bytes + 1;
	m_buff[bytes] = '\n';
	if (read(fh , m_buff , bytes) != bytes) return BAD_RW;
	for (i = 0 ; i < (int)bytes ; i++)
		if (strchr("\t\r\x0" , m_buff[i])) m_buff[i] = *m_separ;
	return SUCCESS;
}

int tokenizer::read_buffer_part(int fh, long bytes)
{
	int err;
	long rbytes = (m_fofs+32000L <= bytes) ? 32000U:(long)(bytes-m_fofs);

	if (rbytes < 1) return BAD_NUMBER;
	lseek(fh , m_fofs , 0);
	err = read_buffer(fh , rbytes);
	if (err != SUCCESS) return err;
	m_fofs += rbytes;
	char* p = &m_buff[m_buffsz-1];
	while (*p != '\n' && p != m_buff-1) {
		p--;
		m_buffsz--;
		m_fofs--;
	}
	return SUCCESS;
}

int tokenizer::read_buffer_frame(int fh, long beg, long end)
{
	int err;
	long rbytes = (end-beg);

	if (rbytes < 1) return BAD_NUMBER;
	lseek(fh , beg , 0);
	err = read_buffer(fh , rbytes);
	return err;
}

int tokenizer::token(char* str, long n)
{
	char* p = str;
	int quit, strb, stre;
	long len = 0;
	register long i = m_ofs;

	strb = FALSE;
	stre = FALSE;
	quit = FALSE;
	while (i < m_buffsz && strchr(m_separ , m_buff[i])) i++;
	if (i < m_buffsz && strchr(m_sticky, m_buff[i])) {
		quit = TRUE;
		p[len++] = m_buff[i++];
	}
	while (!quit) {
		if (i >= m_buffsz) quit = TRUE;
		else if (!strb && (strchr(m_separ,m_buff[i])||strchr(m_sticky,m_buff[i])))
			quit = TRUE;
		else {
			if (m_buff[i] == m_delim[0] && !strb) {
				strb = TRUE;
				if (m_incdelim) p[len++] = m_buff[i];
			} else if (m_buff[i] == m_delim[1] && strb) {
				stre = quit = TRUE;
				if (m_incdelim) p[len++] = m_buff[i];
			} else if (len < n) {
                p[len++] = m_buff[i];
            } else quit = TRUE;
			i++;
		}
	}
	if (len >= n) len = n - 1;
	p[len] = 0;
	if (strb && !stre) return BAD_NAME;
	m_ofs = i;
	return SUCCESS;
}

int tokenizer::compare(const char* s1, const char* s2)
{
	return strcmp(s1 , s2);
}

int tokenizer::preprocess(bool comments, bool defines)
{
	register long i, j;
	bool quit, slrem, mlrem;
	char s[1024], *p;
	int err;

	if (!m_buff) return BAD_ALLOC;

	// remove comments
	if (comments) {
		slrem = mlrem = FALSE;
		for (i = 0 ; i < m_buffsz ; i++)
		{
			if (!(slrem||mlrem) && (m_buff[i]==*m_rem || m_buff[i]==m_rem[2])) {
				if (m_buff[i]==*m_rem && (m_rem[1]==' ' || m_buff[i+1]==m_rem[1]))
					slrem = TRUE;
				else if (m_buff[i]==m_rem[2] && m_buff[i+1]==m_rem[3]) mlrem = TRUE;
			}
			if (slrem) {
				if (m_buff[i] == '\n')
				{
					m_buff[i] = *m_separ;
					slrem = FALSE;
				} else m_buff[i] = *m_separ;
			}
			if (mlrem) {
				if (m_buff[i]==m_rem[4] && m_buff[i+1]==m_rem[5]) {
					m_buff[i] = m_buff[i+1] = *m_separ;
					mlrem = FALSE;
				} else m_buff[i] = *m_separ;
			}
		}
	}

	for (i = 0 ; i < m_buffsz ; i++)
		if (m_buff[i]=='\n' || m_buff[i]=='\r') m_buff[i] = *m_separ;

	_kill_list_entries();

	// make the list
	quit = FALSE;
	m_ofs = 0;
	while (!quit) {
		err = token(s , 1024);
		if (err == SUCCESS) {
			if (*s==0 || m_ofs>m_buffsz) quit = TRUE;
			else if (m_list.append(strdup(s)) == -1) quit = TRUE;
		}
		if (err != SUCCESS) quit = TRUE;
	}

	// replace all defines
	if (defines) {
		char rs[80];
		if (err == SUCCESS) {
			for (i = 0 ; i < m_list.length() ; i++) {
				p = m_list[i];
				if (strcmp(p , "#define") == 0) {
					if (i+2 >= m_list.length()) return BAD_NUMBER;
					strcpy(s , m_list[i+1]);
					strcpy(rs , m_list[i+2]);
					for (j = i+3 ; j < m_list.length() ; j++) {
						p = m_list[j];
						if (!strcmp(s , p)) {
							m_list[j] = strdup(rs);
							delete p;
						}
					}
				}
			}
		}

		// replace all TRUE and FALSE
		if (err == SUCCESS) {
			for (i = 0 ; i < m_list.length() ; i++) {
				p = m_list[i];
				if (!strcmp(p , "TRUE")) {
					m_list[i] = strdup("1");
					delete p;
				}
			}
			for (i = 0 ; i < m_list.length() ; i++) {
				p = m_list[i];
				if (!strcmp(p , "FALSE")) {
					m_list[i] = strdup("0");
					delete p;
				}
			}
		}
	}

	return err;
}

int tokenizer::goto_label(const char* label, long from)
{
	register long i;
	char* p;

	for (i = from ; i < m_list.length() ; i++) {
		p = m_list[i];
		if (!compare(p , label)) {
			if (i+1 >= m_list.length()) return -1;
			return i;
		}
	}
	return -1;
}

bool tokenizer::istext(long p)
{
	char* s = m_list[p];
	long len = strlen(s);
	return bool(s[0]==m_delim[0] && s[(len-1)]==m_delim[1]);
}

int tokenizer::get_text(const char* label, char* ret, long from)
{
	long i = goto_label(label , from);
	if (i == -1) return BAD_INDEX;
	strcpy(ret , m_list[i+1]);
	return SUCCESS;
}

int tokenizer::get_int16(const char* label, short* ret, long from)
{
	long i = goto_label(label , from);
	if (i == -1) return BAD_INDEX;
	*ret = (short)atoi(m_list[i+1]);
	return SUCCESS;
}

int tokenizer::get_int32(const char* label, long* ret, long from)
{
	long i = goto_label(label , from);
	if (i == -1) return BAD_INDEX;
	*ret = (long)atol(m_list[i+1]);
	return SUCCESS;
}

int tokenizer::get_double(const char* label, double* ret, long from)
{
	long i = goto_label(label , from);
	if (i == -1) return BAD_INDEX;
	*ret = (double)atof(m_list[i+1]);
	return SUCCESS;
}

void tokenizer::_kill_list_entries()
{
	for (register long i = 0 ; i < m_list.length() ; i++) {
		char* p = m_list[i];
		if (p) delete p;
	}
	m_list.clear();
}
